
/*
    Interface to v4l2.

    Copyright 2005-2013 by David Santinoli <david@santinoli.com>

    TODO: special treatment for philips and sonix?
*/

#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>
#include <malloc.h>
#include "videodev2.h"
#include <unistd.h>
#include <errno.h>
#include <sys/mman.h>
#include "diag.h"
#include "jpeg.h"
#include "video.h"
#include "video2.h"
#include "caption.h"
#include "yuv2rgb.h"
#include "spca561.h"
#include "parser.h"
#include "pwc-ioctl.h"
#include "sensor_def.h"

#define FOURCC_CHARS(fmt) \
  ((fmt)&0xff),(((fmt)>>8)&0xff),(((fmt)>>16)&0xff),(((fmt)>>24)&0xff)

static int prev_capx=0, prev_capy=0;
static int has_tuners;

static void update_sensors(struct v4l2_queryctrl v2_queryctrl)
{
 int i;
 for(i=0;i<sensors;i++) {
    if(!strcasecmp(sensor[i].name,(char *)v2_queryctrl.name)) {
       diag(SS_VIDEO|DEBUG,"Found device definition for %s\n",v2_queryctrl.name);
       sensor[i].imin=v2_queryctrl.minimum;
       sensor[i].imax=v2_queryctrl.maximum;
       }
    }
}

static void video2_enumerate_menu(struct v4l2_queryctrl *qrctrl)
{
 struct v4l2_querymenu v2_querymenu;
 diag(SS_VIDEO|DEBUG,"Menu items:\n");
 memset(&v2_querymenu,0,sizeof(v2_querymenu));
 v2_querymenu.id=qrctrl->id;
 for (v2_querymenu.index=qrctrl->minimum;v2_querymenu.index<=qrctrl->maximum; v2_querymenu.index++) {
    if(ioctl(cparms.fd,VIDIOC_QUERYMENU,&v2_querymenu)==0)
       diag(SS_VIDEO|DEBUG,"  %s\n",v2_querymenu.name);
    else
       diag(SS_VIDEO|DIERR,"VIDIOC_QUERYMENU");
    }
}

int video2_init(int chan, int mono)
{
 int i, n;
 unsigned int c;
 char *caps[]={
    "video_capture","video_output","video_overlay",0,
    "vbi_capture","vbi_output","sliced_vbi_capture","sliced_vbi_output",
    "rds_capture","video_output_overlay","hw_freq_seek","rds_output",
    0,"video_capture_mplane","video_output_mplane","video_m2m_mplane",
    "tuner","audio","radio","modulator",
    0,0,0,0,
    "readwrite","asyncio","streaming",0,
    0,0,0,"device_caps",
    };
 struct v4l2_capability v2_cap;
 struct v4l2_input v2_input; // XXX should be an array
 struct v4l2_queryctrl v2_queryctrl;
 struct v4l2_fmtdesc v2_fmtd;
 struct v4l2_format v2_fmt;
 struct v4l2_standard standard;
 unsigned int pfmt;

 struct pixf_item {
    unsigned int pixf;
    int depth;
    int type;
    int available;
    } pixformats[]={
  { V4L2_PIX_FMT_BGR24,
    24, 0, 0 },
  { V4L2_PIX_FMT_YUV420,
    12, 1, 0 },
  { V4L2_PIX_FMT_YUYV,
    16, 1, 0 },
  { V4L2_PIX_FMT_UYVY,
    16, 1, 0 },
  { V4L2_PIX_FMT_SBGGR8,
    24, 2, 0 },
  { V4L2_PIX_FMT_SGBRG8,
    24, 2, 0 },
  { V4L2_PIX_FMT_SGRBG8,
    24, 2, 0 },
  { V4L2_PIX_FMT_SRGGB8,
    24, 2, 0 },
  { V4L2_PIX_FMT_SPCA561,
    24, 3, 0 },
  };

 diag(SS_VIDEO|INFO,"Trying V4L2 interface first\n");
 if( (n=ioctl(cparms.fd,VIDIOC_QUERYCAP,&v2_cap)) ) {
    if(errno==EINVAL)
       diag(SS_VIDEO|INFO,"Video device is not V4L2 compatible, reverting to V4L\n");
    else
       diag(SS_VIDEO|INFO|ERRNO,"Error querying capabilities (%d), reverting to V4L",n);
    return 0;
    }

 diag(SS_VIDEO|DEBUG,"Driver:\t%s\n",v2_cap.driver);
 diag(SS_VIDEO|DEBUG,"Card:\t%s\n",v2_cap.card);
 diag(SS_VIDEO|DEBUG,"Bus_info:\t%s\n",v2_cap.bus_info);

 diag(SS_VIDEO|DEBUG,"Capabilities:\t");
 for(c=0;c<32;c++)
    if(v2_cap.capabilities&(1<<c))
       diag(DEBUG|NOTIME,"%s ",(caps[c]?caps[c]:"undefined") );
 diag(DEBUG|NOTIME,"\n");

 if(!(v2_cap.capabilities&V4L2_CAP_VIDEO_CAPTURE)) {
    diag(SS_VIDEO|INFO,"Missing V4L2 capture capability, reverting to V4L\n");
    return 0;
    }

 has_tuners=0;
 for(v2_input.index=0;;v2_input.index++) {
    if(ioctl(cparms.fd,VIDIOC_ENUMINPUT,&v2_input)==-1) break;
    diag(SS_VIDEO|DEBUG,"Video input %d: type %s (\"%s\")\n",
       v2_input.index,
       (v2_input.type==1?"tuner":v2_input.type==2?"composite":"unknown"),
       v2_input.name);
    if(v2_input.type==1) has_tuners=1;
    }

 if(v2_input.index==0) diag(SS_VIDEO|DIERR,"No video inputs available!");

 if(chan>=v2_input.index)
    diag(SS_VIDEO|DIE,"Channel must be in range [0..%d]\n",v2_input.index-1);
 else {
    //diag(SS_VIDEO|DEBUG,"Setting current video input to %d\n",chan);
    if(ioctl(cparms.fd,VIDIOC_S_INPUT,&chan)==-1)
       diag(SS_VIDEO|DIERR,"VIDIOC_S_INPUT");
    }

 memset(&standard,0,sizeof(standard));
 standard.index=0;
 while(ioctl(cparms.fd,VIDIOC_ENUMSTD,&standard)!=-1) {
    diag(SS_VIDEO|DEBUG,"Supported video standard: %lu\n",standard.id);
    standard.index++;
    }

 for(v2_queryctrl.id=V4L2_CID_BASE;v2_queryctrl.id<V4L2_CID_LASTP1;v2_queryctrl.id++) {
    if(ioctl(cparms.fd,VIDIOC_QUERYCTRL,&v2_queryctrl)==0) {
    if(v2_queryctrl.id==0) break; // bttv. Better way to handle it?
       if(v2_queryctrl.flags&V4L2_CTRL_FLAG_DISABLED) continue;
       diag(SS_VIDEO|DEBUG,"control: %s [%d:%d-%d/%d]\n", v2_queryctrl.name, v2_queryctrl.type, v2_queryctrl.minimum, v2_queryctrl.maximum, v2_queryctrl.step);
       // should do this only after checking v2_queryctrl.type
       if(v2_queryctrl.id==V4L2_CID_BRIGHTNESS || v2_queryctrl.id==V4L2_CID_EXPOSURE || v2_queryctrl.id==V4L2_CID_CONTRAST || v2_queryctrl.id==V4L2_CID_GAIN)
          update_sensors(v2_queryctrl);
       if(v2_queryctrl.type==V4L2_CTRL_TYPE_MENU)
           video2_enumerate_menu(&v2_queryctrl);
       }
    else {
       if(errno==EINVAL) continue;
       diag(SS_VIDEO|DIERR,"VIDIOC_QUERYCTRL");
       }
    }

 for(v2_queryctrl.id=V4L2_CID_PRIVATE_BASE;;v2_queryctrl.id++) {
    if(ioctl(cparms.fd,VIDIOC_QUERYCTRL,&v2_queryctrl)==0) {
       if(v2_queryctrl.flags&V4L2_CTRL_FLAG_DISABLED) continue;
       diag(SS_VIDEO|DEBUG,"control: %s [%d:%d-%d/%d]\n", v2_queryctrl.name, v2_queryctrl.type, v2_queryctrl.minimum, v2_queryctrl.maximum, v2_queryctrl.step);
       if(v2_queryctrl.type==V4L2_CTRL_TYPE_MENU)
          video2_enumerate_menu(&v2_queryctrl);
       }
    else {
       if(errno==EINVAL) break;
       diag(SS_VIDEO|DIERR,"VIDIOC_QUERYCTRL");
       }
    }

 memset(&v2_fmtd,0,sizeof(v2_fmtd));
 i=0;
 v2_fmtd.index=i;
 v2_fmtd.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
 while(ioctl(cparms.fd,VIDIOC_ENUM_FMT,&v2_fmtd)!=-1) {
    diag(SS_VIDEO|DEBUG,"available pixformat: %c%c%c%c%s - %s\n",
       FOURCC_CHARS(v2_fmtd.pixelformat),
       (v2_fmtd.flags&V4L2_FMT_FLAG_COMPRESSED)?"*":"",
       v2_fmtd.description);
    v2_fmtd.index=++i;
    }

 for(i=0;i<sizeof(pixformats)/sizeof(struct pixf_item);i++) {

    pfmt=pixformats[i].pixf;
    memset(&v2_fmt,0,sizeof(v2_fmt));
    v2_fmt.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
    v2_fmt.fmt.pix.width=config.capx; 
    v2_fmt.fmt.pix.height=config.capy;
    v2_fmt.fmt.pix.field=V4L2_FIELD_ANY;

    diag(SS_VIDEO|DEBUG,"trying %c%c%c%c...\n",FOURCC_CHARS(pfmt));
    v2_fmt.fmt.pix.pixelformat=pfmt;
    if(ioctl(cparms.fd,VIDIOC_S_FMT,&v2_fmt)==0
       && v2_fmt.fmt.pix.pixelformat==pfmt) {
       diag(SS_VIDEO|DEBUG,"%c%c%c%c supported with capture size %dx%d\n",FOURCC_CHARS(pfmt),v2_fmt.fmt.pix.width,v2_fmt.fmt.pix.height);
       if(v2_fmt.fmt.pix.width!=config.capx || v2_fmt.fmt.pix.height!=config.capy) {
          continue;
          }
       break;
       }
    }
 if(i==sizeof(pixformats)/sizeof(struct pixf_item)) {
    diag(SS_VIDEO|INFO,"No supported palette available for requested size, reverting to V4L\n");
    return 0;
    }

 cparms.palette=pfmt;
 cparms.depth=pixformats[i].depth;
 cparms.palette_type=pixformats[i].type;

 // XXX probe max capture size?
 diag(SS_VIDEO|DEBUG,"Actual capture size:\t(%dx%d)\n",v2_fmt.fmt.pix.width,v2_fmt.fmt.pix.height);
 cparms.capx=v2_fmt.fmt.pix.width;
 cparms.capy=v2_fmt.fmt.pix.height;
 diag(SS_VIDEO|DEBUG,"Image size: %d\n",v2_fmt.fmt.pix.sizeimage);
 diag(SS_VIDEO|DEBUG,"Colorspace: %d\n",v2_fmt.fmt.pix.colorspace);

/*
 if(cparms.palette==V4L2_PIX_FMT_SN9C10X)
    sonix_decompress_init();
*/

 /* image props setup */
 
 return 1;
}


static void video2_init_mmap(void)
{
 struct v4l2_requestbuffers req;
 struct v4l2_buffer buf;
 int n;

 memset(&req,0,sizeof(req));
 req.count=2;
 req.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
 req.memory=V4L2_MEMORY_MMAP;

 if(ioctl(cparms.fd,VIDIOC_REQBUFS,&req)==-1) {
    if(errno==EINVAL)
       diag(SS_VIDEO|DIE,"Video device does not support mmap()\n");
    else
       diag(SS_VIDEO|DIERR,"VIDIOC_REQBUFS");
    }

 if(req.count==0)
    diag(SS_VIDEO|DIE,"Device driver could not allocate mmap() buffers\n");
 diag(SS_VIDEO|DEBUG,"%d mmap() buffers allocated\n",req.count);
 cparms.mmap_n_buffers=req.count;
 cparms.mmap_buffers=malloc(req.count*sizeof(struct mmap_buffer));

 if(!cparms.mmap_buffers)
    diag(SS_VIDEO|DIE,"Could not allocate mmap_buffers array\n");

 for(n=0;n<req.count;n++) {
    memset(&buf,0,sizeof(buf));
    buf.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory=V4L2_MEMORY_MMAP;
    buf.index=n;

    if(ioctl(cparms.fd,VIDIOC_QUERYBUF,&buf)==-1)
       diag(SS_VIDEO|DIERR,"VIDIOC_QUERYBUF");

    cparms.mmap_buffers[n].length=buf.length;
    cparms.mmap_buffers[n].start=mmap(NULL,buf.length,PROT_READ|PROT_WRITE,
       MAP_SHARED,cparms.fd,buf.m.offset);
    if(cparms.mmap_buffers[n].start==MAP_FAILED)
       diag(SS_VIDEO|DIERR,"mmap()");
    else
       diag(SS_VIDEO|DEBUG,"mmap()ed buffer %d (length %d, offset %d)\n",
          n,buf.length,buf.m.offset);
    }
}


/*
   mmap() (or allocate the buffer for read()ing from) the video device.
   Allocate the yuv2rgb buffer if needed.
   Previously mmap()ed/malloc()ed memory is freed before reallocation upon
   change of requested capture size (runtime change of capture size is not
   currently implemented. Changes should take place before restarting capture
   thread).
*/
void video2_alloc_buffers(void)
{
 if(cparms.capx==prev_capx && cparms.capy==prev_capy) return;

 if(prev_capx && prev_capy) {
    video2_free_buffers();
    jpeg_free_rows();
    }

 prev_capx=cparms.capx; prev_capy=cparms.capy;

 /* capture window setup... */

 video2_init_mmap();

 cparms.bufsize=(cparms.depth*cparms.capx*cparms.capy)/8;
 if((cparms.buffer=(unsigned char *)malloc(cparms.bufsize))==0)
    diag(SS_VIDEO|DIE,"malloc() of %d bytes failed",cparms.bufsize);

 if(cparms.palette_type>0) {
    /* actually, this buffer is also used for Bayer and S561 conversion, not only YUV */
    diag(SS_VIDEO|DEBUG,"Allocating yuv->rgb buffer (%d bytes)\n",3*(cparms.capx)*(cparms.capy));
    cparms.yuv2rgbbuf=malloc(3*(cparms.capx)*(cparms.capy));
    /* if YUV->RGB needs a look-up table, initialize it here. */
    }
 else cparms.yuv2rgbbuf=0;

 if(prev_capx && prev_capy)
    jpeg_compress_init(cparms.capx,cparms.capy,
       (cparms.palette_type==0 || cparms.palette_type==3)?cparms.buffer:cparms.yuv2rgbbuf,
       cparms.mono,config.jpeg_quality);
}


void video2_capture_compress(int nojpeg)
{
 struct v4l2_buffer buf;
 int e;

    buf.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory=V4L2_MEMORY_MMAP;
    do {
       e=ioctl(cparms.fd,VIDIOC_DQBUF,&buf);
       } while(e && errno==EINTR);
    if(e) diag(SS_VIDEO|ERR,"VIDIOC_DQBUF");
    if(buf.index>=cparms.mmap_n_buffers)
       diag(SS_VIDEO|ERROR,"Buffer index out of bounds - this is BAD\n");
    /* should use a dedicated buffer for decompression? */
    if(cparms.palette==V4L2_PIX_FMT_SN9C10X) {
#if 0
       if(sonix_decompress(cparms.capx,cparms.capy,cparms.mmap_buffers[buf.index].start,cparms.buffer))
          diag(SS_VIDEO|ERROR,"Sonix frame decompression failed.\n");
#endif
       }
    else memcpy(cparms.buffer,cparms.mmap_buffers[buf.index].start,
       cparms.mmap_buffers[buf.index].length);
    do {
       e=ioctl(cparms.fd,VIDIOC_QBUF,&buf);
       } while(e && errno==EINTR);
    if(e) diag(SS_VIDEO|ERR,"VIDIOC_QBUF");

 if(cparms.palette_type==1) yuv2rgb(cparms.palette,cparms.yuv2rgbbuf,cparms.buffer,cparms.capx,cparms.capy);
 else if(cparms.palette_type==2) {
    bayer2rgb(cparms.palette,cparms.yuv2rgbbuf,cparms.buffer,cparms.capx,cparms.capy);
    }
 else if(cparms.palette_type==3) {
    decode_spca561(cparms.buffer,cparms.yuv2rgbbuf,cparms.capx,cparms.capy);
    /* untested! */
    bayer2rgb(V4L2_PIX_FMT_SGBRG8,cparms.buffer,cparms.yuv2rgbbuf,cparms.capx,cparms.capy);
    }
 else if(cparms.palette==V4L2_PIX_FMT_BGR24) { // BGR->RGB
    int x,y;
    unsigned char c, *p=cparms.buffer, *p1, *p2;
    for(y=0;y<cparms.capy;y++)
       if(config.image_filters&IMAGE_FILTER_FLIP)
          for(x=0,p1=(p+=3*cparms.capx),p2=p1+3*cparms.capx-1;x<cparms.capx/2;x++) {
             c=*(p1); *(p1++)=*(p2); *(p2--)=c;
             c=*(p1); *p1++=*(p2); *p2--=c;
             c=*(p1); *p1++=*p2; *(p2--)=c;
             }
       else
          for(x=0;x<cparms.capx;x++,p+=3) {
             c=*(p+2);
             *(p+2)=*(p);
             *(p)=c;
             }
    }

#if 0
 {
 static int f=0;
 //if(!f) f=open("out.raw",O_WRONLY);
 f=open("out.raw",O_CREAT|O_WRONLY|O_TRUNC,0644);
 write(f,cparms.buffer,3*(cparms.capx)*(cparms.capy));
 close(f);
 }
#endif

 draw_captions(
    (cparms.palette_type==0 || cparms.palette_type==3)? cparms.buffer : cparms.yuv2rgbbuf,
    cparms.capx,cparms.capy);

 if(!nojpeg) jpeg_compress(cparms.capx,cparms.capy);
}


void video2_free_buffers(void)
{
// XXX must revise.
 free(cparms.buffer);
 if(cparms.mbuffer)
    munmap(cparms.mbuffer,cparms.v_mbuf.size);

 if(cparms.palette_type>0) free(cparms.yuv2rgbbuf);
}


void video2_cleanup(void)
{
 video2_free_buffers();
 close(cparms.fd);
}

int video2_norm(char *s)
{
 if(!strcmp(s,"PAL")) return VIDEO_MODE_PAL;
 else if(!strcmp(s,"NTSC")) return VIDEO_MODE_NTSC;
 else if(!strcmp(s,"SECAM")) return VIDEO_MODE_SECAM;
 else if(!strcmp(s,"AUTO")) return VIDEO_MODE_AUTO;
 else return -1;
}

void video2_capture_start(void)
{
 int i;
 enum v4l2_buf_type type;
 struct v4l2_buffer buf;

 for(i=0;i<cparms.mmap_n_buffers;i++) {
    memset(&buf,0,sizeof(buf));
    buf.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory=V4L2_MEMORY_MMAP;
    buf.index=i;
    if(ioctl(cparms.fd,VIDIOC_QBUF,&buf)==-1)
       diag(SS_VIDEO|DIERR,"VIDIOC_QBUF");
    /*diag(SS_VIDEO|DEBUG,"Enqueued buffer %d\n",i);*/
    }

    type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if(ioctl(cparms.fd,VIDIOC_STREAMON,&type)==-1)
         diag(SS_VIDEO|DIERR,"VIDIOC_STREAMON");
}

void video2_capture_stop(void)
{
 enum v4l2_buf_type type;

 type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
 if(ioctl(cparms.fd,VIDIOC_STREAMOFF,&type)==-1)
    diag(SS_VIDEO|DIERR,"VIDIOC_STREAMOFF");
}

void video2_change_param(int par, int val)
{
 struct v4l2_control ctrl;

 ctrl.value=val;
 if(par>=0 && par<=8) {
    switch(par) {
       case 0:
          ctrl.id=V4L2_CID_BRIGHTNESS;
          break;
       case 1:
          ctrl.id=V4L2_CID_CONTRAST;
          break;
       case 2:
          ctrl.id=V4L2_CID_SATURATION;
          break;
       case 7:
          ctrl.id=V4L2_CID_EXPOSURE;
          break;
       case 8:
          ctrl.id=V4L2_CID_GAIN;
          break;
          }
    if(ioctl(cparms.fd,VIDIOC_S_CTRL,&ctrl)) diag(SS_VIDEO|ERR,"VIDIOC_S_CTRL");
    }
}
