
/*
    YUV->RGB conversion.

    Copyright 2000-2013 by David Santinoli <david@santinoli.com>


    Possible improvements:
    - look-up tables
    - use SIMD instructions (MMX etc.)

 
    References:
       http://www.poynton.com/notes/colour_and_gamma/ColorFAQ.txt
       http://cedar.intel.com/cgi-bin/ids.dll/content/content.jsp?cntKey=Legacy::irtm_COLORCON_10809&cntType=IDS_EDITORIAL#6 (dead?)
       ftp://download.intel.com/support/performancetools/libraries/ipp/ippiman.pdf

    * VIDEO_PALETTE_YUV420P aka YUV 4:2:0 (4:2:0 planar)
            YYYYYYYY      N lines
            UUUU          N/2 lines
            VVVV          N/2 lines

    * VIDEO_PALETTE_YUV422 aka YUV 4:2:2 aka YUYV
            YUYVYUYV

    * VIDEO_PALETTE_UYVY
            variation of the above, with U V Y reordering
            UYVYUYVY

    G = 1.164*(Y-16) -0.813*(V-128) -0.391*(U-128)
    B = 1.164*(Y-16)                +2.018*(U-128)
    R = 1.164*(Y-16) +1.596*(V-128)

    (Y,U,V are in fact Y,Cb,Cr)
*/

#include "videodev2.h"
#include "video.h"

#define clip(x) ( (x)<0 ? 0 : ( (x)>255 ? 255 : (x) ) )

void write_rgb(unsigned char **out, int Y, int U, int V)
{
 int R,G,B;
 R=(76284*Y+104595*V)>>16;
 G=(76284*Y -25625*U-53281*V)>>16;
 B=(76284*Y+132252*U)>>16;

 *(*out)=clip(R);
 *(*out+1)=clip(G);
 *(*out+2)=clip(B);
 *out+=3;
}

void yuv2rgb (int palette, unsigned char *out, unsigned char *in, int width, int height)
{
 unsigned char *u=0,*u1,*v=0,*v1;
 int Y=0,U=0,V=0,i,j;

 if(palette==V4L2_PIX_FMT_YUV420) {
 u=in+width*height;
 v=u+(width*height)/4;

 for(i=0;i<height;i++) {
    u1=u;
    v1=v;
    for(j=0;j<width;j++) {
       Y=(*in++)-16;
       if((j&1)==0) {
          U=(*u++)-128;
          V=(*v++)-128;
          }
       write_rgb(&out,Y,U,V);
       }
    if((i&1)==0) { u=u1; v=v1; }
    }
 }
 else if(palette==V4L2_PIX_FMT_YUYV || palette==V4L2_PIX_FMT_UYVY) {
    if(palette==V4L2_PIX_FMT_YUYV) { /* was VIDEO_PALETTE_YUV422 */
       u=in+1; v=in+3;
       }
    else if(palette==V4L2_PIX_FMT_UYVY) { /* was VIDEO_PALETTE_UYVY */
       u=in; v=in+2; in++;
       }
    for(i=0;i<width*height;i++) {
       Y=(*in)-16;
       U=(*u)-128;
       V=(*v)-128;
       write_rgb(&out,Y,U,V);
       in+=2;
       if((i&1)==1) { u+=4; v+=4; }
       }
    }
}


void bayer2rgb(int palette, unsigned char *out, unsigned char *in, int width, int height)
{
 unsigned char *s=in, *d=out, *n[9];
 int x, y, dl, dr, du, dd, m;

 if(palette==V4L2_PIX_FMT_SRGGB8) m=1;
 else if(palette==V4L2_PIX_FMT_SBGGR8) m=2;
 else if(palette==V4L2_PIX_FMT_SGBRG8) m=3;
 else m=0; /* V4L2_PIX_FMT_SGRBG8 */

 for(y=0;y<height;y++) {
    for(x=0;x<width;x++) {
       dl=-1; dr=1; du=-width; dd=width;
       if(x==0) dl=1;
       else if(x==width-1) dr=-1;
       if(y==0) du=width;
       else if(y==height-1) dd=-width;
       n[0]=(s+du+dl);
       n[1]=(s+du);
       n[2]=(s+du+dr);
       n[3]=(s+dl);
       n[4]=(s);
       n[5]=(s+dr);
       n[6]=(s+dd+dl);
       n[7]=(s+dd);
       n[8]=(s+dd+dr);

       switch( ((x%2)+(y%2)*2)^m ) {
          case 0: // R[G]R
             *d++=(*n[3]+*n[5])/2;
             *d++=*n[4];
             *d++=(*n[1]+*n[7])/2;
             break;
          case 1: // G[R]G
             *d++=*n[4];
             *d++=(*n[1]+*n[3]+*n[5]+*n[7])/4;
             *d++=(*n[0]+*n[2]+*n[6]+*n[8])/4;
             break;
          case 2: // G[B]G
             *d++=(*n[0]+*n[2]+*n[6]+*n[8])/4;
             *d++=(*n[1]+*n[3]+*n[5]+*n[7])/4;
             *d++=*n[4];
             break;
          case 3: // B[G]B
             *d++=(*n[1]+*n[7])/2;
             *d++=*n[4];
             *d++=(*n[3]+*n[5])/2;
             break;
          }
       s++;
       }
    }
}
